﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using BasicMaze;

namespace Maize
{
    /// <summary>
    /// Provides the means to draw a game maze.
    /// </summary>
    class MazeDrawing
    {
        //cells that make up the maze
        private Cell[,] _mazeCells;
        //start, finish, dimensions
        private int _xStartPos;
        private int _yStartPos;
        private int _xFinishPos;
        private int _yFinishPos;
        private int _xLength;
        private int _yLength;
        //congratulatory message
        private string _finishString = " You found the corn! ";

        private int _lrIndent = 0;
        /// <summary>
        /// How many pixels from the left of the drawing area to indent the maze.
        /// </summary>
        public int LrIndent
        {
            get { return _lrIndent; }
            set { _lrIndent = value; }
        }
        
        private int _udIndent = 0;
        /// <summary>
        /// How many pixels from the top of the drawing area to indent the maze.
        /// </summary>
        public int UdIndent
        {
            get { return _udIndent; }
            set { _udIndent = value; }
        }

        private int _cellScale = 20; //a default for sanity
        /// <summary>
        /// How many pixels wide or tall the square cell is drawn;
        /// should be the lowest of height / vertical cells, width / horizontal cells.
        /// </summary>
        public int CellScale
        {
            get { return _cellScale; }
            set { _cellScale = value; }
        }        

        /// <summary>
        /// Instantiate a new MazeDrawing - provide everything needed to draw the maze.
        /// </summary>
        /// <param name="cellsIn">2-D Array of BasicMaze Cells.</param>
        /// <param name="xSize">Number of cells in the vertical direction.</param>
        /// <param name="ySize">Number of cells in the horizontal direction.</param>
        /// <param name="xStart">Vertical component of starting position.</param>
        /// <param name="yStart">Horizontal component of starting position.</param>
        /// <param name="xFinish">Vertical component of finishing position.</param>
        /// <param name="yFinish">Horizontal component of finishing position.</param>
        public MazeDrawing(Cell[,] cellsIn, int xSize, int ySize, int xStart, int yStart, int xFinish, int yFinish)
        {
            _mazeCells = cellsIn;
            _xLength = xSize;
            _yLength = ySize;
            _xStartPos = xStart;
            _yStartPos = yStart;
            _xFinishPos = xFinish;
            _yFinishPos = yFinish;
        }

        /// <summary>
        /// Generate a text representation of the maze.
        /// </summary>
        /// <returns>Collection of lines making up a text picture of the maze.</returns>
        public List<string> DrawAsText()
        {
            List <string> mazeText = new List<string>();
            //initialize each line
            for (int i = 0; i < (_xLength * 2) + 1; i++)
            {
                mazeText.Add(string.Empty);
            }

            //draw cells
            for (int i = 0; i < _xLength; i++)
            {
                //upper-lefts
                //  +—
                //  | 
                for (int j = 0; j < _yLength; j++)
                {
                    //if N free
                    if (!_mazeCells[i, j].isWalled((int)BasicMaze.Directions.N))
                    {
                        mazeText[i * 2] = String.Concat(mazeText.ElementAt(i * 2), "+ ");
                    }
                    else
                    {
                        mazeText[i * 2] = String.Concat(mazeText.ElementAt(i * 2), "+—");
                    }
                    //if W free
                    if (!_mazeCells[i, j].isWalled((int)BasicMaze.Directions.W))
                    {
                        mazeText[(i * 2) + 1] = String.Concat(mazeText.ElementAt((i * 2) + 1), " ");
                    }
                    else
                    {
                        mazeText[(i * 2) + 1] = String.Concat(mazeText.ElementAt((i * 2) + 1), "|");
                    }
                    //if start space
                    if (i == _xStartPos && j == _yStartPos)
                    {
                        mazeText[(i * 2) + 1] = String.Concat(mazeText.ElementAt((i * 2) + 1), "S");
                    }
                    //if finish space
                    else if (i == _xFinishPos && j == _yFinishPos)
                    {
                        //I would print a ☺ here, but it's hard to see what it is
                        mazeText[(i * 2) + 1] = String.Concat(mazeText.ElementAt((i * 2) + 1), "F");
                    }
                    //if regular space
                    else
                    {
                        mazeText[(i * 2) + 1] = String.Concat(mazeText.ElementAt((i * 2) + 1), " ");
                    }
                }
                //upper-right
                //  +
                //  |
                mazeText[i * 2] = String.Concat(mazeText.ElementAt(i * 2), "+");
                mazeText[(i * 2) + 1] = String.Concat(mazeText.ElementAt((i * 2) + 1), "|");
            }
            //bottoms
            for (int j = 0; j < _yLength; j++)
            {
                mazeText[_xLength * 2] = String.Concat(mazeText.ElementAt(_xLength * 2), "+—");
            }
            mazeText[_xLength * 2] = String.Concat(mazeText.ElementAt(_xLength * 2), "+");

            return mazeText;
        }

        /// <summary>
        /// Generate a GDI+ drawing of the maze.
        /// </summary>
        /// <param name="e">PaintEventArgs from the target form or panel's draw event.</param>
        /// <param name="winning">Whether or not we are in the "winning" condition.</param>
        public void DrawAsGraphics(PaintEventArgs e, bool winning) 
        {
            //winning rectangle (drawn in "winning" condition)
            int rectHeight;
            int rectWidth;
            int whereUD;
            int whereLR;
            if (_xLength > _yLength)//tall maze
            {
                rectHeight = ((_xLength * _cellScale)+ 1) / 2;
                rectWidth = ((_yLength * _cellScale)+1);
                whereUD = ((_xLength * _cellScale) / 4) + _udIndent;
                whereLR = _lrIndent;
            }
            else if (_yLength > _xLength)//wide maze
            {
                rectHeight = ((_xLength * _cellScale)+ 1);
                rectWidth = ((_yLength * _cellScale)+ 1) / 2;
                whereUD = _udIndent;
                whereLR = ((_yLength * _cellScale) / 4) + _lrIndent;
            }
            else
            {
                rectHeight = ((_xLength * _cellScale)+ 1) / 2;
                rectWidth = ((_yLength * _cellScale)+ 1) / 2;
                whereUD = ((_xLength * _cellScale) / 4) + _udIndent;
                whereLR = ((_yLength * _cellScale) / 4) + _lrIndent;
            }

            //bring in the graphics from the form or panel
            var graphicsObject = e.Graphics;
            
            //size things:
            SizeF mySize;
            Rectangle mazeSpace = new Rectangle(0, 0, _cellScale, _cellScale);//a regular maze space
            Rectangle winSpace = new Rectangle(0, 0, rectWidth, rectHeight);//the winning rectangle
            //size a text character to a maze space
            Font myFont = AppropriateFont(graphicsObject, 1, 1000,
              mazeSpace.Size, "A", new Font(FontFamily.GenericMonospace, 12), out mySize);
            //size the winning text to the winning rectangle
            Font myWinFont = AppropriateFont(graphicsObject, 1, 1000,
              winSpace.Size, _finishString, SystemFonts.DefaultFont, out mySize);
            
            //a place to draw
            PointF myPlace = new PointF();
            //brushes to draw game elements with
            SolidBrush myStartBrush = new SolidBrush(Color.YellowGreen);
            SolidBrush myFinishBrush = new SolidBrush(Color.Yellow);
            SolidBrush myPlayerBrush = new SolidBrush(Color.Green);
            SolidBrush myCrumbBrush = new SolidBrush(Color.Yellow);
            //brushes to draw winning rectangle and winning text
            SolidBrush myWinBrush = new SolidBrush(Color.Black);
            SolidBrush myWinTextBrush = new SolidBrush(Color.White);
            //pen to draw maze walls
            Pen myPen = new Pen(Color.White);
            //pen to draw player marker (instead of using an image)
            //Pen piecePen = new Pen(Color.LightGreen);
            //piecePen.Width = _cellScale / 64;
            
            //fill drawing area with black
            graphicsObject.Clear(Color.Black);

            //draw cells
            for (int i = 0; i < _xLength; i++)
            {
                //upper-lefts
                //  +—
                //  | 
                for (int j = 0; j < _yLength; j++)
                {
                    //if N free
                    if (!_mazeCells[i, j].isWalled((int)BasicMaze.Directions.N))
                    {
                        //don't draw the top line
                    }
                    else
                    {
                        //draw the top line: xstart, ystart, xfinish, yfinish
                        graphicsObject.DrawLine(myPen, (j * _cellScale) + _lrIndent, (i * _cellScale) + _udIndent, (j * _cellScale) + _cellScale + _lrIndent - 1, (i * _cellScale) + _udIndent);
                    }
                    //if W free
                    if (!_mazeCells[i, j].isWalled((int)BasicMaze.Directions.W))
                    {
                        //don't draw the W line
                    }
                    else
                    {
                        //draw the W line
                        graphicsObject.DrawLine(myPen, (j * _cellScale) + _lrIndent, (i * _cellScale) + _udIndent, (j * _cellScale) + _lrIndent, (i * _cellScale) + _cellScale + _udIndent);
                    }

                    //what to do in the space (order of these is important - player comes foremost)
                    //if player space 
                    if (_mazeCells[i,j].PlayerMarker)
                    {
                        //image:
                        myPlace.Y = (i * _cellScale) + _udIndent + 1;
                        myPlace.X = (j * _cellScale) + _lrIndent + 1;
                        //for black on white, use eyk_black
                        graphicsObject.DrawImage(Properties.Resources.eyk_white, myPlace.X, myPlace.Y, _cellScale-1, _cellScale-1);

                        //drawn marker:
                        //myPlace.Y = (i * _cellScale) + _udIndent + (_cellScale / 4);
                        //myPlace.X = (j * _cellScale) + _lrIndent + (_cellScale / 4);
                        //graphicsObject.FillEllipse(myPlayerBrush, myPlace.X, myPlace.Y, (_cellScale / 2), (_cellScale / 2 ));
                        //graphicsObject.DrawEllipse(piecePen, myPlace.X, myPlace.Y, (_cellScale / 2), (_cellScale / 2));
                    }
                    //if finish space
                    else if (i == _xFinishPos && j == _yFinishPos)
                    {
                        myPlace.Y = (i * _cellScale) + _udIndent;
                        myPlace.X = (j * _cellScale) + _lrIndent;
                        graphicsObject.DrawImage(Properties.Resources.ear_of_corn_xpar, myPlace.X, myPlace.Y, _cellScale, _cellScale);
                    }
                    //if start space
                    else if (i == _xStartPos && j == _yStartPos)
                    {
                        myPlace.Y = (i * _cellScale) + _udIndent + (_cellScale / 8);
                        myPlace.X = (j * _cellScale) + _lrIndent + (_cellScale / 8);
                        graphicsObject.DrawString("☻", myFont, myFinishBrush, myPlace);
                    }
                    //if breadcrumb space
                    else if (_mazeCells[i, j].Breadcrumb)
                    {
                        myPlace.Y = (i * _cellScale) + _udIndent + (_cellScale / 2) - (_cellScale / 16);
                        myPlace.X = (j * _cellScale) + _lrIndent + (_cellScale / 2) - (_cellScale / 16);
                        graphicsObject.FillRectangle(myCrumbBrush, myPlace.X, myPlace.Y, (_cellScale / 8), (_cellScale / 8));
                    }
                }
                //upper-right
                //  +
                //  |
                //draw the E lines
                graphicsObject.DrawLine(myPen, (_yLength * _cellScale) + _lrIndent, (i * _cellScale) + _udIndent, (_yLength * _cellScale) + _lrIndent, (i * _cellScale) + _cellScale + _udIndent);
            }
            //bottoms
            for (int j = 0; j < _yLength; j++)
            {
                //draw the S lines
                graphicsObject.DrawLine(myPen, (j * _cellScale) + _lrIndent, (_xLength * _cellScale) + _udIndent, (j * _cellScale) + _cellScale + _lrIndent - 1, (_xLength * _cellScale) + _udIndent);
            }

            //draw the winning rectangle on top of everything, if appropriate
            if (winning)
            {
                //paint rectangle
                graphicsObject.FillRectangle(myWinBrush, whereLR, whereUD, rectWidth, rectHeight);
                //write text (and try to center it a bit)
                myPlace.Y = whereUD;
                if (myFont.GetHeight() < rectHeight)
                {
                    myPlace.Y += (rectHeight - myFont.GetHeight()) / 2;
                }
                myPlace.X = whereLR;
                graphicsObject.DrawString(_finishString, myWinFont, myWinTextBrush, myPlace);
            }
        }

        /// <summary>
        /// Generate a Bitmap of the maze.
        /// </summary>
        /// <param name="imageCellScale">Width of a single (square) cell, in pixels.</param>
        /// <param name="margin">Margin from top and bottom, in pixels.</param>
        /// <returns></returns>
        public Bitmap DrawAsImage(int imageCellScale, int margin)
        {
            //create a bitmap "slate" for the image
            Bitmap mazeImage = new Bitmap((imageCellScale * _yLength + (margin * 2)) + 1, (imageCellScale * _xLength + (margin * 2)) + 1);
            //get its graphics component
            Graphics graphicsObject = Graphics.FromImage(mazeImage);

            //size things:
            SizeF mySize;
            Rectangle mazeSpace = new Rectangle(0, 0, imageCellScale, imageCellScale);//a regular maze space
            //size a character to the maze space
            Font myFont = AppropriateFont(graphicsObject, 1, 1000,
              mazeSpace.Size, "A", new Font(FontFamily.GenericMonospace, 12), out mySize);

            //a place to draw
            PointF myPlace = new PointF();
            //brushes to draw the start and finish spaces
            SolidBrush myStartBrush = new SolidBrush(Color.Red);
            SolidBrush myFinishBrush = new SolidBrush(Color.Blue);
            //a pen to draw the maze walls
            Pen myPen = new Pen(Color.Black);

            //fill with white
            graphicsObject.Clear(Color.White);

            //draw cells
            for (int i = 0; i < _xLength; i++)
            {
                //upper-lefts
                //  +—
                //  | 
                for (int j = 0; j < _yLength; j++)
                {
                    //if N free
                    if (!_mazeCells[i, j].isWalled((int)BasicMaze.Directions.N))
                    {
                        //don't draw the top line
                    }
                    else
                    {
                        //draw the top line: xstart, ystart, xfinish, yfinish
                        graphicsObject.DrawLine(myPen, (j * imageCellScale) + margin, (i * imageCellScale) + margin, (j * imageCellScale) + imageCellScale + margin - 1, (i * imageCellScale) + margin);
                    }
                    //if W free
                    if (!_mazeCells[i, j].isWalled((int)BasicMaze.Directions.W))
                    {
                        //don't draw the W line
                    }
                    else
                    {
                        //draw the W line
                        graphicsObject.DrawLine(myPen, (j * imageCellScale) + margin, (i * imageCellScale) + margin, (j * imageCellScale) + margin, (i * imageCellScale) + imageCellScale + margin);
                    }
                    //if start space
                    if (i == _xStartPos && j == _yStartPos)
                    {
                        myPlace.Y = (i * imageCellScale) + margin + (imageCellScale / 8);
                        myPlace.X = (j * imageCellScale) + margin + (imageCellScale / 8);
                        graphicsObject.DrawString("S", myFont, myStartBrush, myPlace);
                    }
                    //if finish space
                    else if (i == _xFinishPos && j == _yFinishPos)
                    {
                        myPlace.Y = (i * imageCellScale) + margin + (imageCellScale / 8);
                        myPlace.X = (j * imageCellScale) + margin + (imageCellScale / 8);
                        graphicsObject.DrawString("F", myFont, myFinishBrush, myPlace);
                    }
                    //if regular space
                    else
                    {
                        //do nuthin!
                    }
                }
                //upper-right
                //  +
                //  |
                //draw the E lines
                graphicsObject.DrawLine(myPen, (_yLength * imageCellScale) + margin, (i * imageCellScale) + margin, (_yLength * imageCellScale) + margin, (i * imageCellScale) + imageCellScale + margin);
            }
            //bottoms
            for (int j = 0; j < _yLength; j++)
            {
                //draw the S lines
                graphicsObject.DrawLine(myPen, (j * imageCellScale) + margin, (_xLength * imageCellScale) + margin, (j * imageCellScale) + imageCellScale + margin - 1, (_xLength * imageCellScale) + margin);
            }

            //send back the finished Bitmap
            return mazeImage;                
        }

        /// <summary>
        /// Scale a font to fit within particular dimensions
        /// It's a bit crazy-looking since Font instances are immutable
        /// http://tech.pro/tutorial/691/csharp-tutorial-font-scaling
        /// </summary>
        /// <param name="g">Graphics object, needed for its measurement methods.</param>
        /// <param name="minFontSize">Minimum acceptable font size.</param>
        /// <param name="maxFontSize">Maximum acceptable font size.</param>
        /// <param name="layoutSize">The space you have to say something in.</param>
        /// <param name="s">What you want to say in the space.</param>
        /// <param name="f">The font you want to use.</param>
        /// <param name="extent">Size object to save to.</param>
        /// <returns></returns>
        public static Font AppropriateFont(Graphics g, float minFontSize,
            float maxFontSize, Size layoutSize, string s, Font f, out SizeF extent)
        {
            //if min and max acceptable are the same, we're basically done in a few lines
            if (maxFontSize == minFontSize)
            {
                f = new Font(f.FontFamily, minFontSize, f.Style);
            }

            //measure size with current font (so we can fire return)
            extent = g.MeasureString(s, f);

            //if max was less than min, we're definitely done!
            if (maxFontSize <= minFontSize)
            {
                return f;
            }

            //calculate the ratio of the available size over the size of the string
            float hRatio = layoutSize.Height / extent.Height;
            float wRatio = layoutSize.Width / extent.Width;
            //above 1 means there is space left, and we can make the font bigger
            //below 1 means that the font is too big and needs to be made smaller
            float ratio = (hRatio < wRatio) ? hRatio : wRatio;

            //multiply the current font size by that ratio to get the correct font size for the space
            float newSize = f.Size * ratio;

            //above or below max/min provided?
            if (newSize < minFontSize)
            {
                newSize = minFontSize;
            }
            else if (newSize > maxFontSize)
            {
                newSize = maxFontSize;
            }

            //create the new font
            f = new Font(f.FontFamily, newSize, f.Style);
            //measure the string with that font
            extent = g.MeasureString(s, f);
            //and here ya go
            return f;
         }
    }
}
